package com.sksamuel.jmxc

import javax.management.MBeanServerConnection
import javax.management.remote.{JMXConnectorFactory, JMXServiceURL, JMXConnector}
import scala.collection.JavaConverters._
import sun.jvmstat.monitor.{MonitoredVmUtil, VmIdentifier, MonitoredHost}
import sun.management.ConnectorAddressLink

/** @author Stephen Samuel */
class JmxClient(conn: MBeanServerConnection) {

    def count: Int = conn.getMBeanCount
    def domains: Seq[JmxDomain] = conn.getDomains.map(new JmxDomain(_, conn))
    def domain(name: String): Option[JmxDomain] = domains.find(_.name == name)
}

object JmxClient {

    /**
     * Create a new client connected to the  address specified by the given jmx service url.
     *
     * @param connectorAddress
     *
     * @return a new remote attached JmxClient
     */
    def apply(connectorAddress: String): JmxClient = apply(connectorAddress, null, null)

    /**
     * Create a new client connected to the remote address specified by the given jmx service url
     * and validated with the provided credentials
     *
     * @param connectorAddress
     *
     * @param username the username
     *
     * @param password the password
     *
     * @return a new remote attached JmxClient
     */
    def apply(connectorAddress: String, username: String, password: String): JmxClient = {
        val env = scala.collection.mutable.Map.empty[String, Array[String]]
        if (username != null)
            env.put(JMXConnector.CREDENTIALS, Array(username, password))
        _createClient(connectorAddress, env.toMap)
    }

    /**
     * Create a new client connected to the local process identified by the vm display name
     *
     * @param displayName the display name of the VM
     *
     * @return a connected client or None if no such display name was found
     */
    def fromVmDisplayName(displayName: String): Option[JmxClient] =
        JmxClient.localAttachables.find(_.name == displayName).map(arg => apply(arg.address))

    /**
     * Attempt to create a client for any local process that has a domain that matches the given domain name.
     * This is very useful when you want to include jmcx in a deployable, and then find the local VM that contains the
     * beans for your project at runtime.
     *
     * @param name the name of the required domain
     *
     * @return optional jmx client for a local process that had a domain matching the param.
     */
    def fromDomainName(name: String): Option[JmxClient] = {
        localClients.find(arg => {
            arg.domain(name) match {
                case None => false
                case Some(d) => true
            }
        })
    }

    def localClients = JmxClient.localAttachables.map(arg => apply(arg.address))

    def _createClient(connectorAddress: String, env: Map[String, Array[String]]) = {
        val serviceAddress = new JMXServiceURL(connectorAddress)
        val connector = JMXConnectorFactory.connect(serviceAddress, env.asJava)
        val conn = connector.getMBeanServerConnection
        new JmxClient(conn)
    }

    /**
     * Returns the process ids of all local vms
     *
     * @return
     */
    def localVmPids: Seq[Int] = {
        val host = MonitoredHost.getMonitoredHost(null.asInstanceOf[String])
        host.activeVms().asScala.map(_.toInt).toList
    }

    def localVms: Seq[LocalVm] = {
        val host = MonitoredHost.getMonitoredHost(null.asInstanceOf[String])
        host.activeVms().asScala.map(pid => {
            val mvm = host.getMonitoredVm(new VmIdentifier(pid.toString))
            val name = MonitoredVmUtil.commandLine(mvm)
            val attachable = MonitoredVmUtil.isAttachable(mvm)
            val address = ConnectorAddressLink.importFrom(pid)
            mvm.detach()
            LocalVm(name, attachable, address, pid)
        }).toList
    }

    def localAttachables: Seq[LocalVm] = localVms.filter(arg => arg.attachable && arg.address != null)

    case class LocalVm(name: String, attachable: Boolean, address: String, pid: Int)
}
